#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include "functions.h"
#include "interface.h"
#include "support.h"


struct t_struct {
   int peg[3][NUM_STONES]; /* peg[0][x] is on the left peg[2][x] is on the right.     */
                           /* The values of the numbers stored coorespond to the size */
                           /*     of the stone. The largest stone is 6 and the        */
                           /*     smallest is 1. peg[x][0] is the top of the peg, and */
                           /*     peg[x][5] is the bottom.                            */
   int lifted_peg,  /* The peg that the user is trying to move a stone from.   */
                    /*     a -1 shows that no peg is lifted.                   */
       num_moves;   /* This is the total number of moves taken                 */
};
struct t_struct t_s, t_d;
static struct t_struct *tower_data,
                       *tower_data_backup;

static GtkWidget *drawing_area = NULL,
                 *status_bar;

static GdkPixmap *stone_xpm[NUM_STONES];
static GdkPixmap *peg_top_xpm;
static GdkPixmap *peg_xpm;
static GdkPixmap *pixmap = NULL; /* a double buffer for the drawing_area */


/********************************
 *  main_window menu callbacks  *
 ********************************/
void on_restart_menu_activate(GtkMenuItem *menuitem, gpointer user_data)
{
   init_tower_data();
   init_drawing_area(lookup_widget(GTK_WIDGET(menuitem), "main_window"));
}


void on_quit_menu_activate(GtkMenuItem *menuitem, gpointer user_data)
{
   /* maybe add one of those "Are you sure.." dialogs here */
   gtk_main_quit();
}


void on_undo_menu_activate(GtkMenuItem *menuitem, gpointer user_data)
{
   undo_backup(SET_UNDO_BACKUP);
}


void on_instructions_menu_activate(GtkMenuItem *menuitem, gpointer user_data)
{
   GtkWidget *instructions_dialog;
   instructions_dialog = create_instructions_dialog();
   gtk_widget_show(instructions_dialog);
}


void on_about_menu_activate(GtkMenuItem *menuitem, gpointer user_data)
{
   GtkWidget *about_dialog;
   about_dialog = create_about_dialog();
   gtk_widget_show(about_dialog);
}


/**********************
 *  Dialog callbacks  *
 **********************/
void on_about_button_released(GtkButton *button, gpointer user_data)
{
   gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(button)));
}


void on_instructions_button_released(GtkButton *button, gpointer user_data)
{
   gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(button)));
}


void on_game_over_quit_button_released(GtkButton *button, gpointer user_data)
{
   gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(button)));
   gtk_main_quit();
}


void on_game_over_retry_button_released(GtkButton *button, gpointer user_data)
{
   init_tower_data();
   /* use drawing area because its a static widget in the main_window */
   init_drawing_area(lookup_widget(drawing_area, "main_window"));

   gtk_widget_destroy(gtk_widget_get_toplevel(GTK_WIDGET(button)));
}


/****************************
 *  drawing_area callbacks  *
 ****************************/
gboolean on_drawingarea_configure_event(GtkWidget *widget,
                                        GdkEventConfigure *event,
                                        gpointer user_data)
{ /* this function is just a 'just in case' sort of thing, it  *
   * isn't needed as of now. The main window isn't resizable   */
   if (pixmap)
      gdk_pixmap_unref(pixmap);

   pixmap = gdk_pixmap_new(widget->window,
                           widget->allocation.width,
                           widget->allocation.height,
                           -1);
   gdk_draw_rectangle (pixmap,
                       widget->style->white_gc,
                       TRUE,
                       0, 0,
                       widget->allocation.width,
                       widget->allocation.height);

   return TRUE;
}


gboolean on_drawingarea_expose_event(GtkWidget *widget,
                                     GdkEventExpose *event,
                                     gpointer user_data)
{
   /* redraw the area that just got uncovered */
   gdk_draw_pixmap(widget->window,
                   widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
                   pixmap,
                   event->area.x, event->area.y,
                   event->area.x, event->area.y,
                   event->area.width, event->area.height);

  return FALSE;
}


gboolean on_drawing_area_button_press_event(GtkWidget *widget,
                                            GdkEventButton *event,
                                            gpointer user_data)
{
   int i = 0,
       j = 0,
       clicked_peg = (int)(((event->x)+1)/PIXMAP_WIDTH); /* returns 0,1, or 2 depending on where *
                                                          * the mouse button was pushed.         */
                                                         /* Also, I add 1 to event->x just in    *
                                                          * case the user clicked where x=0.     */
   char buff[40];

   /* find first(top) occupied spot on the peg that was clicked. */
   /* A 6 indicates it is empty.                                 */
   while((tower_data->peg[clicked_peg][i] == 0) && (i++ != NUM_STONES-1)){}

   if(tower_data->lifted_peg != -1)         /* if there is already a stone selected to move  */
   {
      if(tower_data->lifted_peg != clicked_peg)/* if we're not trying to set it back on the same peg */
      {
         while((tower_data->peg[tower_data->lifted_peg][j] == 0) && (j++ != NUM_STONES-1)){} /* get the position of the lifted stone */

         if(tower_data->peg[clicked_peg][i] > tower_data->peg[tower_data->lifted_peg][j] ||
            i == NUM_STONES) /* the top stone on the clicked peg is smaller than the one we're trying to move. *
                              * ...or the peg is empty. In other words, we're good to go.                      */
         {
            undo_backup(GET_UNDO_BACKUP);

            tower_data->peg[clicked_peg][i-1] = tower_data->peg[tower_data->lifted_peg][j];
            tower_data->peg[tower_data->lifted_peg][j] = 0;
            tower_data->num_moves++;

            /* update the statusbar */
            g_snprintf(buff, 40, "Number of moves taken so far: %d", tower_data->num_moves);
            gtk_statusbar_pop(GTK_STATUSBAR(status_bar), 1);
            gtk_statusbar_push(GTK_STATUSBAR(status_bar), 1, buff);
         }
      }
      tower_data->lifted_peg = -1;
   }
   else if(tower_data->peg[clicked_peg][NUM_STONES-1] != 0)
      tower_data->lifted_peg = clicked_peg; /* there isn't one lifted, so pick one up    */

   draw_screen();

   if((tower_data->peg[0][0] == 1) || (tower_data->peg[2][0])) /* we have a winner */
      you_win(widget);

   return TRUE;
}


/*********************
 *  Misc. functions  *
 *********************/
void init_tower_data()
{
   short int i;

   tower_data = &t_d; /* give tower_data something to point to   */

   for(i = 0; i < NUM_STONES; i++)
   {
      tower_data->peg[0][i] = 0;   /* an empty peg                */
      tower_data->peg[1][i] = i+1; /* put smallest (1) peg on top *
                                    * and largest (6) on bottom   */
      tower_data->peg[2][i] = 0;   /* another empty peg           */
   }
   tower_data->lifted_peg = -1;
   tower_data->num_moves = 0;
}


void init_drawing_area(GtkWidget *main_window)
{
   short int i;
   GdkBitmap *mask;
   GtkStyle *style;

   style = gtk_widget_get_style(main_window);

   /* give the static global drawing_area something to point to.      *
    * 'drawingarea' was the name as created by interface.c from glade */
   drawing_area = lookup_widget(main_window, "drawingarea");

   /* the status bar isn't really the drawingarea, but we'll set it up *
      here anyway. (its already created in interface.c)                */
   status_bar = lookup_widget(main_window, "statusbar");

   /* initial text for the statusbar */
   gtk_statusbar_pop(GTK_STATUSBAR(status_bar), 1);
   gtk_statusbar_push(GTK_STATUSBAR(status_bar), 1, "Number of moves taken so far: 0");

   peg_xpm = gdk_pixmap_create_from_xpm(drawing_area->window, &mask,
                                        &style->bg[GTK_STATE_NORMAL],
                                        PACKAGE_DATA_DIR "/pixmaps/peg.xpm" );
   if(!peg_xpm)
      g_error("Could not open file: peg.xpm\nin directory: %s", PACKAGE_DATA_DIR);

   peg_top_xpm = gdk_pixmap_create_from_xpm(drawing_area->window, &mask,
                                            &style->bg[GTK_STATE_NORMAL],
                                            PACKAGE_DATA_DIR "/pixmaps/peg_top.xpm");
      if(!peg_top_xpm)
         g_error("Could not open file: /pixmaps/peg_top.xpm\nin directory: %s", PACKAGE_DATA_DIR);

   stone_xpm[0] = gdk_pixmap_create_from_xpm(main_window->window, &mask,
                                             &style->bg[GTK_STATE_NORMAL],
                                             PACKAGE_DATA_DIR "/pixmaps/stone_0.xpm");
   stone_xpm[1] = gdk_pixmap_create_from_xpm(drawing_area->window, &mask,
                                             &style->bg[GTK_STATE_NORMAL],
                                             PACKAGE_DATA_DIR "/pixmaps/stone_1.xpm");
   stone_xpm[2] = gdk_pixmap_create_from_xpm(drawing_area->window, &mask,
                                             &style->bg[GTK_STATE_NORMAL],
                                             PACKAGE_DATA_DIR "/pixmaps/stone_2.xpm");
   stone_xpm[3] = gdk_pixmap_create_from_xpm(drawing_area->window, &mask,
                                             &style->bg[GTK_STATE_NORMAL],
                                             PACKAGE_DATA_DIR "/pixmaps/stone_3.xpm");
   stone_xpm[4] = gdk_pixmap_create_from_xpm(drawing_area->window, &mask,
                                             &style->bg[GTK_STATE_NORMAL],
                                             PACKAGE_DATA_DIR "/pixmaps/stone_4.xpm");
   stone_xpm[5] = gdk_pixmap_create_from_xpm(drawing_area->window, &mask,
                                             &style->bg[GTK_STATE_NORMAL],
                                             PACKAGE_DATA_DIR "/pixmaps/stone_5.xpm");
      for(i = 1; i < NUM_STONES; i++)
         if(!stone_xpm[i])
            g_error("Could not open file: /pixmaps/stone_%d.xpm\nin directory: %s", i, PACKAGE_DATA_DIR);

   draw_screen();
}


void draw_screen()
{
   short int i, j;

   /* draw the tops of the pegs */
   gdk_draw_pixmap(pixmap,
                   drawing_area->style->fg_gc[GTK_WIDGET_STATE(drawing_area)],
                   peg_top_xpm,
                   0, 0,
                   0, 0,
                   PIXMAP_WIDTH, PIXMAP_HEIGHT);
   gdk_draw_pixmap(pixmap,
                   drawing_area->style->fg_gc[GTK_WIDGET_STATE(drawing_area)],
                   peg_top_xpm,
                   0, 0,
                   PIXMAP_WIDTH, 0,
                   PIXMAP_WIDTH, PIXMAP_HEIGHT);
   gdk_draw_pixmap(pixmap,
                   drawing_area->style->fg_gc[GTK_WIDGET_STATE(drawing_area)],
                   peg_top_xpm,
                   0, 0,
                   2*PIXMAP_WIDTH, 0,
                   PIXMAP_WIDTH, PIXMAP_HEIGHT);

   /* fill in the rest of the drawing_area */
   for(i = 0; i < NUM_STONES; i++)
      for(j = 0; j < 3; j++)
      {
         /* when peg_x[i]==0, the peg on that spot is empty */
         if(tower_data->peg[j][i] == 0)
            gdk_draw_pixmap(pixmap,
                            drawing_area->style->fg_gc[GTK_WIDGET_STATE(drawing_area)],
                            peg_xpm,
                            0,0,
                            j*PIXMAP_WIDTH, (i+1)*PIXMAP_HEIGHT,
                            PIXMAP_WIDTH, PIXMAP_HEIGHT);
         else /* otherwise use a stone of the relevent size */
            gdk_draw_pixmap(pixmap,
                            drawing_area->style->fg_gc[GTK_WIDGET_STATE(drawing_area)],
                            stone_xpm[(tower_data->peg[j][i])-1], /* we -1 because a peg_x[i] value cooresponds *
                                                                   * to the size of the stone, where size 0     *
                                                                   * indicate empty, and 1 is the smallest.     */
                            0, 0,
                            j*PIXMAP_WIDTH, (i+1)*PIXMAP_HEIGHT,
                            PIXMAP_WIDTH, PIXMAP_HEIGHT);
      }

   /* refresh */
   gdk_draw_pixmap(drawing_area->window,
                   drawing_area->style->fg_gc[GTK_WIDGET_STATE(drawing_area)],
                   pixmap,
                   0, 0,
                   0, 0,
                   3*PIXMAP_WIDTH, NUM_STONES*PIXMAP_HEIGHT+PIXMAP_HEIGHT); /* add the extra PIXMAP_HEIGHT for the peg_top_xpm */

   if(tower_data->lifted_peg != -1) /* animate the lifting of a stone */
   {
      i = 0;
      while((tower_data->peg[tower_data->lifted_peg][i] == 0) && (i++ != NUM_STONES-1)){}
      for(j = 0; j < LIFT_HEIGHT; j++)
      {
         /* draw the peg to sit behind the gap */
         gdk_draw_pixmap(pixmap,
                         drawing_area->style->fg_gc[GTK_WIDGET_STATE(drawing_area)],
                         peg_xpm,
                         0, 0,
                         tower_data->lifted_peg*PIXMAP_WIDTH, (i+1)*PIXMAP_HEIGHT,
                         PIXMAP_WIDTH, PIXMAP_HEIGHT);

         /* move the stone up one pixel */
         gdk_draw_pixmap(pixmap,
                         drawing_area->style->fg_gc[GTK_WIDGET_STATE(drawing_area)],
                         stone_xpm[tower_data->peg[tower_data->lifted_peg][i]-1],
                         0, 0,
                         tower_data->lifted_peg*PIXMAP_WIDTH, (i+1)*PIXMAP_HEIGHT-j,
                         PIXMAP_WIDTH, PIXMAP_HEIGHT);
         /* refresh */
         gdk_draw_pixmap(drawing_area->window,
                drawing_area->style->fg_gc[GTK_WIDGET_STATE(drawing_area)],
                pixmap,
                tower_data->lifted_peg*PIXMAP_WIDTH, (i+1)*PIXMAP_HEIGHT-j,
                tower_data->lifted_peg*PIXMAP_WIDTH, (i+1)*PIXMAP_HEIGHT-j,
                PIXMAP_WIDTH, PIXMAP_HEIGHT+j);
      }
   }
}


void you_win()
{
   GtkWidget *game_over_dialog;
   char *message;

   if(tower_data->num_moves == 63)
      message = "\nSlap mah fro!! (jk)\n\nYou finished with a perfect 63 moves!!!\nSorry though, no flashy winner screen yet =[\n";
   else if(tower_data->num_moves < 72)
      message = "\nCongratulations.\n\nNot too shabby.";
   else
      message = "\nWell, you finished.\n\nBut that is about all you can say.\nI haven't seen such a terrible\nscore in quite some time.\n";

   game_over_dialog = create_game_over_dialog(message);
   gtk_widget_show(game_over_dialog);
}


/* I'd much rather return the struct t_struct, but its  *
 * not possible as the on_undo_menu_activate() is       *
 * called by gtk_main().                                */
void undo_backup(get_undo_backup)
{
   static short int first_loop = 1;
   int i, j;
   char buff[40];

   if(first_loop)
   {
      first_loop = 0;
      tower_data_backup = &t_s;
      tower_data_backup->num_moves = 0;
      tower_data_backup->lifted_peg = -1;
   }

   /* In functions.h you can see the defines for GET/SET_UNDO_BACKUP */
   if(get_undo_backup)
      for(i = 0; i < 3; i++)
         for(j = 0; j < NUM_STONES; j++)
         {
            tower_data_backup->peg[i][j] = tower_data->peg[i][j];
            tower_data_backup->num_moves = tower_data->num_moves;
            tower_data_backup->lifted_peg = -1;
         }
   else
   {
      for(i = 0; i < 3; i++)
         for(j = 0; j < NUM_STONES; j++)
         {
            tower_data->peg[i][j] = tower_data_backup->peg[i][j];
            tower_data->num_moves = tower_data_backup->num_moves;
            tower_data->lifted_peg = -1;
            g_snprintf(buff, 40, "Number of moves taken so far: %d", tower_data->num_moves);
            gtk_statusbar_pop(GTK_STATUSBAR(status_bar), 1);
            gtk_statusbar_push(GTK_STATUSBAR(status_bar), 1, buff);
         }
      draw_screen();
   }
}